body {
  margin: 0;
  overflow: hidden;
}
canvas {
  margin: 0;
}
div {
  margin: 0;
}

.btn {
  min-height: 2em;
  height: 2em;
}

.navbar {
  position: absolute;
  min-height: 2.1em;
  height: 2.1em;
}

.loading_bar_data {
  position: absolute;
  top: 2.1em;
  width: 0%;
  height: 3px;
  background-color: #009fff;
}

.particle_menu {
  position: absolute;
  background: transparent;
  top: 45%;
  max-height: 40%;
  max-width: 20%;
  display: none;
}

.dropdown:hover > .dropdown-menu {
  display: block;
}

#all_toggle_options {
  position: absolute;
  width: 8em;
  top: 10%;
  right: 0%;
  padding: 1em;
}

.toggle_options {
  padding-bottom: 0.5em;
}

#scale_bar_div {
  position: absolute;
  height: 5em;
  bottom: 5%;
  left: 2%;
}

#scale_bar {
  width: 10em;
  height: 0.5em;
  visibility: hidden;
}

#scale_bar_text {
  visibility: hidden;
}

.evd_image {
  position: absolute;
  top: 50%;
  left: 50%;
  -webkit-transform: translate(-50%, -50%);
  transform: translate(-50%, -50%);
  border: 3px solid rgba(255, 255, 255, 0.5);
}

.quit_message {
  position: absolute;
  top: 50%;
  left: 50%;
  -webkit-transform: translate(-50%, -50%);
  transform: translate(-50%, -50%);
  width: 12em;
  display: none;
  opacity: 0;
}

.citations {
  position: absolute;
  bottom: 0;
  left: 0;
  width: 100%;
  font-size: 0.8em;
  text-align: center;
}

.mc_truth {
  position: absolute;
  top: 5%;
  left: 0;
  width: 100%;
  text-align: center;
}

input:active,
input:focus,
input:hover,
input:visited,
input.active,
input.focus,
input.hover,
input.visited,
button:active,
button:focus,
button:hover,
button:visited,
button.active,
button.focus,
button.hover,
button.visited {
  box-shadow: unset !important;
  outline: 0 !important;
}
